#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HASHSIZE 100

char *strsave(char *s) /*save string s somewhere*/
{
char *p;

p = (char *)malloc(strlen(s)+1);

if(! p) return 0;
strcpy(p,s);

return p;
}

static char temp[5000];
struct alist{
char *name;
char *def;
int uflag;
struct alist *next;
};
static struct alist *hashtab[HASHSIZE];/*pointer table*/

int hash(s)/*form hash value for string s */
char *s;
{
int hashval;
for(hashval=0;*s!='\0';)hashval+= *s++;
/* sum of ascii value of characters divived by tablesize */
return(hashval % HASHSIZE);
}

int getmark(s)
char *s;
{
struct alist *ap;
for(ap=hashtab[hash(s)];ap!=0;ap=ap->next)
	{
	if(strcmp(s,ap->name)==0)return(ap->uflag);
	}
return(-1);
}

void putmark(s)
char *s;
{
struct alist *ap;
for(ap=hashtab[hash(s)];ap!=0;ap=ap->next)
if(strcmp(s,ap->name)==0)
	{
	ap->uflag=1;
	}
}

struct alist *lookup(s)  /*look for s in hashtab */
char *s;
{
struct alist *ap;
for(ap=hashtab[hash(s)];ap!=0;ap=ap->next)/*ap points to next entry*/
if(strcmp(s,ap->name)==0)return(ap);/*found it (the name pointed to by ap)*/
return(0);/*not found*/
}

struct alist *install(name,def,mode)/*put (name.def) in hashtab*/
char *name,*def;
int mode;/*0=full symbol table, 1=restricted memory use, only one def*/
{
struct alist *ap,*lookup();
int hashval;
if((ap=lookup(name))==0) /*not found*/
	{
	ap=(struct alist *)malloc(sizeof(*ap));/*make space for new structure*/
	if(ap==0)return(0);/*if no space return 0*/
	if((ap->name=strsave(name))==0)return(0);/*if no space for name*/
	hashval=hash(ap->name);/*calc hashvalue of new name*/
	ap->next=hashtab[hashval];/*ap->next now points to prev. entry struct*/
	hashtab[hashval]=ap;/*table now points to present entry*/
/*note that new structure was entered at bottom of chain of structures*/
	strcpy(temp,def);
	ap->uflag=0;/*reset the used flag on new entry*/
	}
else /*already there*/
	{
	if(mode==0)
		{
		strcpy(temp,ap->def);
		strcat(temp," ");
		strcat(temp,def);
		}
	else
		{
		strcpy(temp,"");
		}
	free(ap->def);
	}
if((ap->def=strsave(temp))==0)return(0);
return(ap);
}

void prilist(FILE *f, int pce)
{
struct alist *ap;
char s[20];
int i;
fprintf(f,"\nList of addresses and their use, the format is:\n\
address\treferred to by\tby\tand so on.\n");
for(i=0;i<pce;i++)
	{
	strcpy(s,"");
	sprintf(s,"%04x",i);
	ap=lookup(s);
	if(ap!=0)
		{
		fprintf(f,"%s: %d %s\n",ap->name,ap->uflag,ap->def);
		}
	}
fprintf(f,"\nEnd symbol table\n");
}


int tosymtab(ia,rfl,prc)/*calculate destination addresses and store in hash table*/
int ia[],rfl;
unsigned int prc;
{
unsigned int argval;
int argtyp;
char sr[20],sv[20],sw[20];

#define	JPDEST	1
#define	JRDEST	2
#define	CLDEST	3
#define	VRBYTE	4
#define	VRWORD	5
#define	CSTADD	6
#define DJDEST	7

argtyp=0;
if((ia[0]==0x32) || (ia[0]==0x3a))
	{
	argtyp=VRBYTE;
	argval=ia[1]+256*ia[2];
	}
else if((ia[0]==0x22) || (ia[0]==0x2a))
	{
	argtyp=VRWORD;
	argval=ia[1]+256*ia[2];
	}
else if(ia[0]==0xed)
	{
	if((ia[1]==0x43)||(ia[1]==0x4b)||(ia[1]==0x53)||(ia[1]==0x5b)\
	||(ia[1]==0x73)||(ia[1]==0x7b))
		{
		argtyp=VRWORD;
		argval=ia[2]+256*ia[3];
		}
	}
else if((ia[0]==0xdd)||(ia[0]==0xfd))
	{
	if((ia[1]==0x22)||(ia[1]==0x2a)) // BUGFIX Ric.Hohne 17/7/2008
		{
		argtyp=VRWORD;
		argval=ia[2]+256*ia[3];
		}
	if(ia[1]==0x21)
		{
		argtyp=CSTADD;
		argval=ia[2]+256*ia[3];
		}
	}
else if((ia[0]==0xda)||(ia[0]==0xfa)||(ia[0]==0xd2)||(ia[0]==0xc3)\
||(ia[0]==0xc2)||(ia[0]==0xf2)||(ia[0]==0xea)||(ia[0]==0xe2)||(ia[0]==0xca))
	{
	argtyp=JPDEST;
	argval=ia[1]+256*ia[2];
	}

else if((ia[0]==0x38)||(ia[0]==0x18)||(ia[0]==0x30)||(ia[0]==0x20)||(ia[0]==0x28))
	{
	argtyp=JRDEST;
	argval=prc+ia[1]+2;
	if(ia[1]>=128)argval-=256;/*signbit*/
	}
else if((ia[0]==0xdc)||(ia[0]==0xfc)||(ia[0]==0xd4)||(ia[0]==0xcd)||(ia[0]==0xc4)\
||(ia[0]==0xf4)||(ia[0]==0xf4)||(ia[0]==0xec)||(ia[0]==0xe4)||(ia[0]==0xcc))
	{
	argtyp=CLDEST;
	argval=ia[1]+256*ia[2];
	}
else if((ia[0]==0x31)||(ia[0]==0x01)||(ia[0]==0x11)||(ia[0]==0x21))
	{
	argtyp=CSTADD;
	argval=ia[1]+256*ia[2];
	}
else if(ia[0]==0x10)
	{
	argtyp=DJDEST;
	argval=prc+ia[1]+2;
	if(ia[1]>=128)argval-=256;
	}
/*
get argtype, if argval already exists for this type, print prc as source,
else store argval in list and print prc as source.
*/
if(argtyp!=0)
	{
/* prc is always printed in list*/
	switch(argtyp)
		{
		case JPDEST:
				strcpy(sr,"jp");
				break;
		case JRDEST:
				strcpy(sr,"jr");
				break;
		case CLDEST:
				strcpy(sr,"cl");
				break;
		case VRBYTE:
				strcpy(sr,"vb");
				break;
		case VRWORD:
				strcpy(sr,"vw");
				break;
		case CSTADD:
				strcpy(sr,"ac");
				break;
		case DJDEST:
				strcpy(sr,"dj");
				break;
		}/*end switch*/
	strcat(sr," ");
	sprintf(sw,"%04x",prc);
	if(rfl)strcpy(sr,sw);
	else strcat(sr,sw);
	sprintf(sv,"%04x",argval);
/*	printf("\nINSTALLING  sv=%s  sr=%s\n",sv,sr);
*/
	if(install(sv,sr,rfl)==0)return(0);/*no space*/
	}/*end argtyp !=0*/
return(1);/*ok*/
}
